#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Blast-proof utilities.
'''

#FIXME: Consider offshoring this entire subpackage into a new independent
#Python project codenamed "Blast-proof utilities." PiPI project and package
#names deriving from this codename include:
#
#* "blapu". No existing PyPI packages. No prominent Google hits. Sorta ideal? My
#  own quandry is... well, let's face it: "Blapu." Really? That's awful.
#* "blaut". No existing PyPI packages. Many prominent Google hits. Less ideal?
#* "blauti". No existing PyPI packages. No prominent Google hits. *MOST IDEAL?*
#* "blapru". No existing PyPI packages. No prominent Google hits. Less ideal?
#* "blastu". No existing PyPI packages. Extremely prominent Google hits
#  pertaining to a technology festival hosted in Bravil, however. Non-ideal?
#
#I'm partial to "blauti," as of this writing. It just works. Only 5,000 Google
#hits effectively guarantees uniqueness. It also reads and sounds apropos.
#FIXME: Oh, yeah. We got this:
#
#* "brutil". Speaks for itself in waves of languid ludicrousness, doesn't it?
